"""Parsers to convert uncontrolled cell grids into representations of StarTable blocks.

parse_blocks() emits a stream of blocks objects.
This in principle allows early abort of reads as well as generic postprocessing (
as discussed in store-module docstring).

parse_blocks() switches between different parsers depending on the StarTable block type:
- Metadata
- Directive
- Table
- Template (not yet implemented)

For each of these:

- The intended input is a two-dimensional "cell grid" i.e. a sequence of rows, with each row
  being a sequence of values (where a "sequence" is usually a list, but can also be e.g. a tuple).
  Rows need not be of equal length; only the relevant portion of a cell grid will be parsed
  depending on the relevant block type.

- The output is a representation of the StarTable block, either as:
  - A pdtable-style block object e.g. Table
  - A JSON-like data structure ready for serialization via e.g. json.dump()  (only implemented for
    table blocks at this stage); or
  - The original, raw cell grid, in case the user wants to do some low-level processing.

"""
import re
from typing import Sequence, Optional, Tuple, Any, Iterable

import pandas as pd

from .fixer import ParseFixer
from .columns import parse_column
from ... import frame
from pdtable.io._json import to_json_serializable, JsonData, JsonDataPrecursor
from ...auxiliary import MetadataBlock, Directive
from pdtable import Table
from pdtable import BlockType, BlockIterator
from ...table_metadata import TableOriginCSV, TableMetadata

# Typing alias: 2D grid of cells with rows and cols. Intended indexing: cell_grid[row][col]
CellGrid = Sequence[Sequence]

VALID_PARSING_OUTPUT_TYPES = {"pdtable", "jsondata", "cellgrid"}


def make_metadata_block(cells: CellGrid, origin: Optional[str] = None, **_) -> MetadataBlock:
    mb = MetadataBlock(origin)
    for row in cells:
        if len(row) > 1 and row[0] is not None:
            key_field = row[0].strip()
            if len(key_field) > 0 and key_field[-1] == ":":
                mb[key_field[:-1]] = row[1].strip()
    return mb


def make_directive(cells: CellGrid, origin: Optional[str] = None, **_) -> Directive:
    name = cells[0][0][3:]
    directive_lines = [row[0] for row in cells[1:]]
    return Directive(name, directive_lines, origin)


def default_fixer(**kwargs):
    """ Determine if user has supplied custom fixer
        Else return default ParseFixer() instance.
    """
    fixer = kwargs.get("fixer")
    if fixer is not None:
        if type(fixer) is type:
            # It's a class, not an instance. Make an instance here.
            fixer = kwargs["fixer"]()
        else:
            assert isinstance(fixer, ParseFixer)
    else:
        fixer = ParseFixer()
    assert fixer is not None
    fixer.origin = kwargs.get("origin")
    return fixer


def make_table_json_precursor(cells: CellGrid, **kwargs) -> JsonDataPrecursor:
    """Parses cell grid into a JSON-like data structure but with some non-JSON-native values

    Parses cell grid to a JSON-like data structure of nested "objects" (dict), "arrays" (list),
    and values, including values with types that map 1:1 to JSON-native types, as well as some
    value types that don't directly map JSON types.

    This JSON data "precursor" can then be sent for further processing:
    - Parsing to pdtable-style Table block object
    - Conversion to a "pure" JSON data object in which all values are of JSON-native types.
    """

    table_name = cells[0][0][2:]

    fixer = default_fixer(**kwargs)
    fixer.table_name = table_name

    # internally hold destinations as json-compatible dict
    destinations = {dest: None for dest in cells[1][0].strip().split(" ")}

    # handle multiple columns w. same name
    col_names_raw = cells[2]
    column_names = preprocess_column_names(col_names_raw, fixer)
    fixer.TableColumNames = column_names  # TODO typo... no effect... intended behaviour?

    n_col = len(column_names)
    units = cells[3][:n_col]
    units = [el.strip() for el in units]

    column_data = [ll[:n_col] for ll in cells[4:]]
    column_data = [list(col) for col in column_data]

    # ensure all data columns are populated
    for irow, row in enumerate(column_data):
        if len(row) < n_col:
            fix_row = fixer.fix_missing_rows_in_column_data(
                row=irow, row_data=row, num_columns=n_col
            )
            column_data[irow] = fix_row

    # build dictionary of columns iteratively to allow meaningful error messages
    columns = {}
    for name, unit, values in zip(column_names, units, zip(*column_data)):
        try:
            fixer.column_name = name
            columns[name] = parse_column(unit, values, fixer)
        except ValueError as e:
            raise ValueError(
                f"Unable to parse value in column '{name}' of table '{table_name}' as '{unit}'"
            ) from e

    fixer.report()

    return {
        "name": table_name,
        "columns": columns,
        "units": units,
        "destinations": destinations,
        "origin": kwargs.get("origin"),
    }


def make_table(cells: CellGrid, origin: Optional[TableOriginCSV] = None, **kwargs) -> Table:
    """Parses cell grid into a pdtable-style Table block object."""
    json_precursor = make_table_json_precursor(cells, origin=origin, **kwargs)
    return Table(
        frame.make_table_dataframe(
            pd.DataFrame(json_precursor["columns"]),
            units=json_precursor["units"],
            table_metadata=TableMetadata(
                name=json_precursor["name"],
                destinations=set(json_precursor["destinations"].keys()),
                origin=json_precursor["origin"],
            ),
        )
    )


def make_table_json_data(cells: CellGrid, origin, **kwargs) -> JsonData:
    """Parses cell grid into a JSON-ready data structure."""
    impure_json = make_table_json_precursor(cells, origin=origin, **kwargs)
    # attach unit directly to individual column
    units = impure_json["units"]
    del impure_json["units"]  # replaced by "unit" field in columns
    del impure_json["origin"]  # not relevant for json_data
    columns = {}
    for cname, unit in zip(impure_json["columns"].keys(), units):
        columns[cname] = {"unit": unit, "values": impure_json["columns"][cname]}
    impure_json["columns"] = columns
    return to_json_serializable(impure_json)


def make_block(
    block_type: BlockType, cells: CellGrid, origin, **kwargs
) -> Tuple[Optional[BlockType], Optional[Any]]:
    """Dispatches cell grid to the proper parser, depending on block type and desired output type"""
    block_name = ""
    if block_type == BlockType.METADATA:
        factory = make_metadata_block
    elif block_type == BlockType.DIRECTIVE:
        factory = make_directive
    elif block_type == BlockType.TABLE:
        block_name = cells[0][0][2:]
        to = kwargs.get("to")
        if to == "cellgrid":
            factory = lambda c, *_, **__: c  # Regurgitate the unprocessed cell grid  # noqa:E731
        elif to == "jsondata":
            factory = make_table_json_data
        else:
            factory = make_table
    else:
        factory = None

    block_filter = kwargs.get("filter")
    if block_filter:
        assert callable(block_filter)
        if not block_filter(block_type, block_name):
            return None, None

    return block_type, cells if factory is None else factory(cells, origin, **kwargs)


# Regex that matches valid block start markers
_re_block_marker = re.compile(
    r"^("  # Marker must start exactly at start of cell
    r"(?<!\*)(\*\*\*?)(?!\*)"  # **table or ***directive but not ****undefined
    r"|"
    r"((?<!:):{1,3}(?!:))[^:]*\s*$"  # :col, ::table, :::file but not ::::undefined, :ambiguous:
    r"|"
    r"([^:]+:)\s*$"  # metadata:  but not :ambiguous:
    r")"
)
# $1 = Valid block start marker
# $2 = **Table / ***Directive
# $3 = :Template
# $4 = Metadata:


def parse_blocks(cell_rows: Iterable[Sequence], **kwargs) -> BlockIterator:
    """Parses blocks from a single sheet as rows of cells.

    Takes an iterable of cell rows and parses it into blocks.

    Args:
        cell_rows: Iterable of cell rows, where each row is a sequence of cells.
    kwargs:
        origin: A thing.
        fixer: Also a thing, but different.
        to: generate Table of this type ("pdtable", "jsondata", "cellgrid")

    Yields:
        Blocks.
    """

    # Unpack, pre-process, validate, and repack the kwargs for downstream use
    to = kwargs.get("to")
    if to is None:
        kwargs["to"] = to = "pdtable"
    elif to not in VALID_PARSING_OUTPUT_TYPES:
        raise ValueError(
            f"Unknown parsing output type; expected one of {VALID_PARSING_OUTPUT_TYPES}.", to
        )

    origin = kwargs["origin"] if "origin" in kwargs else "stream"

    fixer = default_fixer(**kwargs)
    kwargs["fixer"] = fixer  # use in make_block
    fixer.reset_fixes()
    fixer.origin = origin

    def is_blank(cell):
        """
        True if first cell is empty
        """
        return cell is None or (isinstance(cell, str) and not cell.strip())

    cell_grid = []

    state = BlockType.METADATA
    next_state = None
    this_block_1st_row = 0
    for row_number_0based, row in enumerate(cell_rows):
        #  print(f"parse_blocks: {state} {row[0] if len(row) > 0 else ' (empty) '}")
        if row is None or len(row) == 0 or is_blank(row[0]):
            if state != BlockType.BLANK:
                next_state = BlockType.BLANK
            else:
                continue
        elif isinstance(row[0], str):
            # possible token
            mm = _re_block_marker.match(row[0])
            if mm is None:  # TBC (PEP 572)
                cell_grid.append(row)
                continue

            if mm.group(1) is not None:
                if mm.group(1) == "**":
                    next_state = BlockType.TABLE
                elif mm.group(1) == "***":
                    next_state = BlockType.DIRECTIVE
                elif mm.group(4) is not None:
                    if state == BlockType.METADATA:
                        cell_grid.append(row)
                        continue
                    else:
                        next_state = BlockType.BLANK
                else:
                    next_state = BlockType.TEMPLATE_ROW
        else:
            # binary (excel &c.)
            cell_grid.append(row)
            continue

        if next_state is not None:
            # Current block has ended. Emit it.
            if len(cell_grid) > 0:
                kwargs["origin"] = TableOriginCSV(origin, this_block_1st_row)
                block_type, block = make_block(state, cell_grid, **kwargs)
                if block_type is not None:
                    yield block_type, block
            # TODO augment TableOriginCSV with one tailored for Excel
            cell_grid = []
            state = next_state
            next_state = None
            if state != BlockType.BLANK:
                cell_grid.append(row)
            elif len(row) > 0:
                #  emit non-empty lines, comments &c. as BLANK
                if len(row) == 1 and is_blank(row[0]):
                    continue
                cell_grid.append(row)

    if cell_grid:
        # Block ended with EOF. Emit it.
        kwargs["origin"] = TableOriginCSV(origin, this_block_1st_row)
        block_type, block = make_block(state, cell_grid, **kwargs)
        if block_type is not None:
            yield block_type, block


def preprocess_column_names(col_names_raw: Sequence[str], fixer: ParseFixer):
    """
       handle known issues in column_names
    """
    n_names_col = len(col_names_raw)
    for el in reversed(col_names_raw):
        if el is not None and len(el) > 0:
            break
        n_names_col -= 1

    # handle multiple columns w. same name
    column_names = []
    cnames_all = [el.strip() for el in col_names_raw[:n_names_col]]
    names = {}
    for col, cname in enumerate(cnames_all):
        if cname not in names and len(cname) > 0:
            names[cname] = 0
            column_names.append(cname)
        else:
            fixer.column_name = col
            fixer.TableColumNames = column_names  # so far
            if len(cname) == 0:
                cname = fixer.fix_missing_column_name(input_columns=column_names)
            elif cname in names:
                cname = fixer.fix_duplicate_column_name(cname, input_columns=column_names)
            assert cname not in names
            names[cname] = 0
            column_names.append(cname)
    return column_names



import numpy as np
import pandas as pd
import datetime as dt
from textwrap import dedent

from pdtable import Table, TableBundle, BlockType
from pdtable import ParseFixer
from ....io.parsers.blocks import (
    make_metadata_block,
    make_directive,
    make_table,
    parse_blocks,
)


def test_make_metadata_block():
    cells = [
        [cell.strip() for cell in line.split(";")]
        for line in dedent(
            """\
    author:;XYODA;
    purpose:;Save the galaxy
    """
        )
        .strip()
        .split("\n")
    ]
    ml = make_metadata_block(cells, ";")
    assert ml["author"] == "XYODA"
    assert ml["purpose"] == "Save the galaxy"


def test_make_directive():
    cells = [
        [cell.strip() for cell in line.split(";")]
        for line in dedent(
            """\
    ***foo;
    bar;
    baz;
    """
        )
        .strip()
        .split("\n")
    ]
    d = make_directive(cells)
    assert d.name == "foo"
    assert d.lines == ["bar", "baz"]


def test_make_table():
    lines = [
        ["**foo", None, None, None],
        ["all", None, None, None],
        ["place", "distance", "ETA", "is_hot"],
        ["text", "km", "datetime", "onoff"],
        ["home", 0.0, dt.datetime(2020, 8, 4, 8, 0, 0), 1],
        ["work", 1.0, dt.datetime(2020, 8, 4, 9, 0, 0), 0],
        ["beach", 2.0, dt.datetime(2020, 8, 4, 17, 0, 0), 1],
        ["wonderland", "-", "-", "FALSE"],
    ]

    t = make_table(lines)

    assert t.name == "foo"
    assert set(t.metadata.destinations) == {"all"}
    assert t.column_names == ["place", "distance", "ETA", "is_hot"]
    assert t.units == ["text", "km", "datetime", "onoff"]

    df = pd.DataFrame(
        [
            ["home", 0.0, dt.datetime(2020, 8, 4, 8, 0, 0), True],
            ["work", 1.0, dt.datetime(2020, 8, 4, 9, 0, 0), False],
            ["beach", 2.0, dt.datetime(2020, 8, 4, 17, 0, 0), True],
            ["wonderland", np.nan, np.nan, False],
        ],
        columns=["place", "distance", "ETA", "is_hot"],
    )
    pd.testing.assert_frame_equal(t.df, df)


def test_make_table__with_backslashes():
    cells = [
        [cell.strip() for cell in line.split(";")]
        for line in dedent(
            r"""
**input_files_derived;
all;
file_bytes;file_date;file_name;has_table;
-;text;text;onoff;
15373;20190516T104445;PISA_Library\results\check_Soil_Plastic_ULS1-PISA_C1.csv;1;
15326;20190516T104445;PISA_Library\results\check_Soil_Plastic_ULS1-PISA_C2.csv;1;
"""
        )
        .strip()
        .split("\n")
    ]

    t = make_table(cells).df
    assert t.file_bytes[0] == 15373

    tt = Table(t)
    assert tt.name == "input_files_derived"
    assert set(tt.metadata.destinations) == {"all"}
    assert tt.units == ["-", "text", "text", "onoff"]


def test_make_table__parses_onoff_column():
    cells = [
        [cell.strip() for cell in line.split(";")]
        for line in dedent(
            r"""
    **input_files_derived;
    all;
    file_bytes;file_date;has_table;
    -;text;onoff;
    15373;a;0;
    15326;b;1;
    """
        )
        .strip()
        .split("\n")
    ]

    table_df = make_table(cells).df
    assert table_df.file_bytes[0] == 15373
    assert not table_df.has_table[0]
    assert table_df.has_table[1]
    tt = Table(table_df)
    assert tt.name == "input_files_derived"
    assert set(tt.metadata.destinations) == {"all"}
    assert tt.units == ["-", "text", "onoff"]


def test_make_table__no_trailing_sep():
    cells = [
        [cell.strip() for cell in line.split(";")]
        for line in dedent(
            r"""
        **foo
        all
        column;pct;dash;mm;
        text;%;-;mm;
        bar;10;10;10;
        """
        )
        .strip()
        .split("\n")
    ]

    t = make_table(cells).df
    assert t.column[0] == "bar"
    assert t.dash[0] == 10


def test_parse_blocks():
    cell_rows = [
        line.split(";")
        for line in dedent(
            """\
        author: ;XYODA     ;
        purpose:;Save the galaxy

        ***gunk
        grok
        jiggyjag

        **foo
        all
        column;pct;dash;mm;
        text;%;-;mm;
        bar;10;10;10;

        ::;Table foo describes
        ;the fooness of things
        :.column;Column is a column in foo

        **input_files_derived;
        all;
        file_bytes;file_date;has_table;
        -;text;onoff;
        15373;a;0;
        15326;b;1;
        """
        )
        .strip()
        .split("\n")
    ]

    blocks = list(parse_blocks(cell_rows))

    metadata_blocks = [b for t, b in blocks if t == BlockType.METADATA]
    assert len(metadata_blocks) == 1
    mb = metadata_blocks[0]
    assert mb["author"] == "XYODA"
    assert mb["purpose"] == "Save the galaxy"

    directives = [b for t, b in blocks if t == BlockType.DIRECTIVE]
    assert len(directives) == 1
    d = directives[0]
    assert d.name == "gunk"
    assert d.lines == ["grok", "jiggyjag"]

    tabs = [b for t, b in blocks if t == BlockType.TABLE]
    assert len(tabs) == 2
    t = tabs[0]
    assert t.name == "foo"
    assert t.df["column"].iloc[0] == "bar"

    # Bundle
    table_bundle = TableBundle(parse_blocks(cell_rows), as_dataframe=True)
    assert table_bundle.foo.column.values[0] == "bar"
    assert table_bundle.foo.dash.values[0] == 10


def test_parse_blocks__filters_correctly():
    """ Unit test
        Verify that misc. float input are handled consistently
    """
    # Make some input data cell rows
    # fmt: off
    cell_rows = [
        ["**incl_1"],
        ["dst2"],
        ["species" , "a3"  , "a2"  , "a1"  , "a4"  ],
        ["text"    , "-"   , "-"   , "-"   , "-"   ],
        ["chicken" , 1     , 2     , 3     , 4     ],
        [ ],
        ["**xcl_1"],
        ["dst2"],
        ["species" , "a3"  , "a2"  , "a1"  , "a4"  ],
        ["text"    , "-"   , "-"   , "-"   , "-"   ],
        ["chicken" , 1     , 2     , 3     , 4     ],
        [ ],
        ["**incl_2"],
        ["dst2"],
        ["species" , "a3"  , "a2"  , "a1"  , "a4"  ],
        ["text"    , "-"   , "-"   , "-"   , "-"   ],
        ["chicken" , 1     , 2     , 3     , 4     ],
        [ ],
        ["**xcl_2"],
        ["dst2"],
        ["species" , "a3"  , "a2"  , "a1"  , "a4"  ],
        ["text"    , "-"   , "-"   , "-"   , "-"   ],
        ["chicken" , 1     , 2     , 3     , 4     ],
        [ ],
    ]
    # fmt: on

    # Make some interesting block filter
    def table_whose_name_starts_with_i(tp: BlockType, tn: str) -> bool:
        """Only keeps table blocks whose name starts with 'i', filters out everything else"""
        return tp == BlockType.TABLE and tn[0] == "i"

    # Filter and parse the blocks
    tables = [b for _, b in parse_blocks(cell_rows, filter=table_whose_name_starts_with_i)]

    assert len(tables) == 2
    assert tables[0].name == "incl_1"


def test_parse_blocks__filters_correctly_for_different_output_types():
    """ API test, parse_blocks:
        Verify that parse_blocks filters correctly and parses correctly to different output types
    """

    # Make some input data cell rows
    # fmt: off
    cell_rows = [
        ["**ignore_me"],
        ["dst2"],
        ["species" , "a3"  , "a2"  , "a1"  , "a4"  ],
        ["text"    , "-"   , "-"   , "-"   , "-"   ],
        ["chicken" , 1     , 2     , 3     , 4     ],
        [ ],
        ["**ignore_me_2"],
        ["dst2"],
        ["species" , "a3"  , "a2"  , "a1"  , "a4"  ],
        ["text"    , "-"   , "-"   , "-"   , "-"   ],
        ["chicken" , 1     , 2     , 3     , 4     ],
        [ ],
        ["**keep_me!"],
        ["dst2"],
        ["species" , "a3"  , "a2"  , "a1"  , "a4"  ],
        ["text"    , "-"   , "-"   , "-"   , "-"   ],
        ["chicken" , 5     , 6     , 3.14    , 8   ],
        [ ],
    ]
    # fmt: on

    # Make some interesting block filter
    def table_asking_to_be_kept(tp: BlockType, tn: str) -> bool:
        """Only keeps table blocks named 'keep_me!', filters out everything else"""
        return tp == BlockType.TABLE and tn == "keep_me!"

    # Parse the input, filtering it using that filter, and returning different output types
    cell_grid = [
        b for _, b in parse_blocks(cell_rows, filter=table_asking_to_be_kept, to="cellgrid")
    ]
    json_data = [
        b for _, b in parse_blocks(cell_rows, filter=table_asking_to_be_kept, to="jsondata")
    ]
    tables = [b for _, b in parse_blocks(cell_rows, filter=table_asking_to_be_kept, to="pdtable")]

    # Expected result from filtering and parsing to those different output types?
    assert len(cell_grid) == 1
    assert isinstance(cell_grid[0], list)
    assert cell_grid[0][4][3] == 3.14

    assert len(json_data) == 1
    assert isinstance(json_data[0], dict)
    assert json_data[0]["columns"]["a1"]["values"][0] == 3.14

    assert len(tables) == 1
    assert isinstance(tables[0], Table)
    assert tables[0].df["a1"][0] == 3.14


def test_filter_read_csv():
    """ API test, read_csv:
        Verify that read_csv returns correct values for to in
           {"pdtable", "jsondata", "cellgrid"}
    """
    #
    # TTT TBD: test "to", "filter" on all high-level API's
    # read_bundle &c.
    pass


def test_read_csv_compatible1():
    """
      test_read_csv_compatible

      handle '-' in cells
      handle leading and trailing wsp
    """

    # fmt off
    cell_rows = [
        ["**test_input"],
        ["all"],
        ["numerical", "dates", "onoffs"],
        ["-", "datetime", "onoff"],
        [123, "08/07/2020", 0],
        [123, "08-07-2020", 1],
        [123, "08-07-2020", 1],
        [1.23, None, None],
        [1.23, None, None],
        [1.23, None, None],
        [-1.23, None, None],
        [1.23, None, None],
    ]
    # fmt on
    fix = ParseFixer()
    fix.stop_on_errors = False
    fix._called_from_test = True
    table_bundle = TableBundle(parse_blocks(cell_rows, fixer=fix), as_dataframe=True)
    assert table_bundle

    assert not table_bundle.test_input.onoffs[0]
    assert table_bundle.test_input.onoffs[1]
    assert table_bundle.test_input.onoffs[2]
    for idx in range(0, 3):
        assert table_bundle.test_input.dates[idx].year == 2020
        assert table_bundle.test_input.dates[idx].month == 7
        assert table_bundle.test_input.dates[idx].day == 8

    for idx in range(0, 3):
        assert table_bundle.test_input.numerical[idx] == 123

    assert table_bundle.test_input.numerical[3] == 1.23
    assert table_bundle.test_input.numerical[5] == 1.23
    assert table_bundle.test_input.numerical[7] == 1.23
    assert table_bundle.test_input.numerical[6] == -1.23


def test_read_csv_compatible2():
    """
      test_read_csv_compatible2

      handle leading and trailing wsp in column_name, unit
    """

    cell_rows = [
        line.split(";")
        for line in dedent(
            r"""
    **test_input;
    all;
    numerical ; dates; onoffs ;
     - ; datetime;onoff ;
    123;08/07/2020;0;
    """
        )
        .strip()
        .split("\n")
    ]

    table_bundle = TableBundle(parse_blocks(cell_rows), as_dataframe=True)
    assert table_bundle

    assert not table_bundle.test_input.onoffs[0]
    assert table_bundle.test_input.dates[0].year == 2020
    assert table_bundle.test_input.numerical[0] == 123


def test_parse_blocks__block_types():
    """ Unit test
        Verify that block types are  are handled consistently
    """
    # fmt: off
    cell_rows = [
        ["author: ", "XYODA"                       ],
        ["purpose:", "Save the galaxy"             ],
        ["**tab_1"                                 ],
        ["all" ,                                   ],
        ["species" , "a3"  , "a2"  , "a1"  , "a4"  ],
        ["text"    , "-"   , "-"   , "-"   , "-"   ],
        ["chicken" , 1     , 2     , 3     , 4     ],
        [ ],                                           # term. by newline
        ["**tab_2"                                 ],
        ["all" ,                                   ],
        ["species" , "a3"  , "a2"  , "a1"  , "a4"  ],
        ["text"    , "-"   , "-"   , "-"   , "-"   ],
        ["chicken" , 1     , 2     , 3     , 4     ],
        ["**tab_3" ,                               ],  # term. by table
        ["all" ,                                   ],
        ["species" , "a3"  , "a2"  , "a1"  , "a4"  ],
        ["text"    , "-"   , "-"   , "-"   , "-"   ],
        ["chicken" , 1     , 2     , 3     , 4     ],
        ["***foo"],                                    # term. by directive
        ["bar"],
        ["baz"],
        [":template", "whatnot?"],
        [],
        ["# extra lines 1", "check type"],             # BLANK 1
        ["# extra lines 2"],
        ["**tab_4"                                 ],
        ["all"                                     ],
        ["species" , "a3"  , "a2"  , "a1"  , "a4"  ],
        ["text"    , "-"   , "-"   , "-"   , "-"   ],
        ["chicken" , 1     , 2     , 3     , 4     ],
        ["meta-test1:", "test meta not in header"  ],  # BLANK 2
        ["meta-test2:", "test meta not in header"  ],  # BLANK 3
        ["**tab_5"                                 ],
        ["all"                                     ],
        ["species" , "a3"  , "a2"  , "a1"  , "a4"  ],
        ["text"    , "-"   , "-"   , "-"   , "-"   ],
        ["chicken" , 1     , 2     , 3     , 4     ],
        ["***foo2"],
        ["bar"],
        ["baz"],
    ]
    # fmt: on

    seen = {}
    for ty, block in parse_blocks(cell_rows, to="cellgrid"):
        #  print(f"\n-oOo- {ty} {block}")
        if seen.get(ty) is None:
            seen[ty] = []
        seen[ty].append(block)

    assert len(seen[BlockType.METADATA]) == 1
    assert len(seen[BlockType.TABLE]) == 5
    assert len(seen[BlockType.DIRECTIVE]) == 2
    assert len(seen[BlockType.TEMPLATE_ROW]) == 1
    assert len(seen[BlockType.BLANK]) == 3


def test_parse_blocks__test_demo():
    """ Unit test
        Verify that block types are  are handled consistently
    """
    # fmt: off
    cell_rows = [
       ["author:" ,"XYODA"                     ],
       ["purpose:","Save the galaxy"           ],
       [                                       ],
       ["***gunk"                              ],
       ["grok"                                 ],
       ["jiggyjag"                             ],
       [                                       ],
       ["**places",                            ],
       ["all"                                  ],
       ["place","distance","ETA","is_hot"      ],
       ["text","km","datetime","onoff"         ],
       ["home",0.0,"2020-08-04 08:00:00",1     ],
       ["work",1.0,"2020-08-04 09:00:00",0     ],
       ["beach",2.0,"2020-08-04 17:00:00",1    ],
       [                                       ],
       ["**farm_animals"                       ],
       ["your_farm my_farm other_farm"         ],
       ["species","n_legs","avg_weight"        ],
       ["text","-","kg"                        ],
       ["chicken",2,2                          ],
       ["pig",4,89                             ],
       ["cow",4,200                            ],
       ["unicorn",4,None                       ],
     ]
    # fmt: on

    seen = {}
    for ty, block in parse_blocks(cell_rows, to="cellgrid"):
        # print(f"\n-oOo- {ty} {block}")
        if seen.get(ty) is None:
            seen[ty] = []
        seen[ty].append(block)

    assert len(seen.get(BlockType.METADATA)) == 1
    assert len(seen.get(BlockType.TABLE)) == 2
    assert len(seen.get(BlockType.DIRECTIVE)) == 1
    assert seen.get(BlockType.TEMPLATE_ROW) is None
    assert seen.get(BlockType.BLANK) is None



import json
import os
from pathlib import Path
from textwrap import dedent

import numpy as np
import pandas as pd

from pdtable import BlockType, ParseFixer
from pdtable.io import json_data_to_table, table_to_json_data
from pdtable.io._json import to_json_serializable
from pdtable.io.parsers import parse_blocks
from pdtable.io.parsers.blocks import make_table


class custom_test_fixer(ParseFixer):
    def __init__(self):
        ParseFixer.__init__(self)
        self.stop_on_errors = False
        self._called_from_test = True


def input_dir() -> Path:
    return Path(__file__).parent / "input/with_errors"


def test_json_pdtable():
    """ ensure dict-obj to pdtable conversion
        compare to target created w. read_stream_csv
    """
    cell_rows = [
        line.split(";")
        for line in dedent(
            """\
        **farm_types1;;;
        your_farm my_farm farms_galore;;;
        species;  num;  flt;    log;
        text;       -;   kg;  onoff;
        chicken;    2;    3;      1;
        pig;        4;   39;      0;
        goat;       4;    -;      1;
        zybra;      4;    -;      0;
        cow;      NaN;  200;      1;
        goose;      2;    9;      0;
        """
        )
        .strip()
        .split("\n")
    ]
    pandas_pdtab = None
    # with io.StringIO(csv_src) as fh:
    g = parse_blocks(cell_rows, **{"origin": '"types1.csv" row 1'}, fixer=custom_test_fixer)
    for tp, tab in g:
        pandas_pdtab = tab
    # fmt: off
    table_json_data = {
      "name": "farm_types1",
      "columns": {
         "species": {"unit": "text",
                     "values": ["chicken", "pig", "goat", "zybra", "cow", "goose"]},
         "num": {"unit": "-",
                 "values": [2.0, 4.0, 4.0, 4.0, None, 2.0]},
         "flt": {"unit": "kg",
                 "values": [3.0, 39.0, None, None, 200.0, 9.0]},
         "log": {"unit": "onoff",
                 "values": [True, False, True, False, True, False]}
      },
      "destinations": {"your_farm": None, "my_farm": None, "farms_galore": None}
    }
    # fmt: on

    json_pdtab = json_data_to_table(table_json_data, fixer=custom_test_fixer)
    assert pandas_pdtab.equals(json_pdtab)


def test_json_data_to_pdtable():
    """ ensure dict-obj to pdtable conversion
        compare to target created w. make_table(List[List]])
    """
    # Make a table using the cell grid parser
    lines_target = [
        ["**farm_types1"],
        ["your_farm my_farm farms_galore"],
        ["species", "num", "flt", "log"],
        ["text", "-", "kg", "onoff"],
        ["chicken", 2, 3, 1],
        ["pig", 4, 39, 0],
        ["goat", 4, None, 1],
        ["zybra", 4, None, 0],
        ["cow", None, 200, 1],
        ["goose", 2, 9, 0],
    ]

    table_from_cell_grid = make_table(lines_target, fixer=custom_test_fixer)

    # Make an identical table, but starting from JSON

    # fmt: off
    table_json_data = {
      "name": "farm_types1",
      "columns": {
         "species": {"unit": "text",
                     "values": ["chicken", "pig", "goat", "zybra", "cow", "goose"]},
         "num": {"unit": "-",
                 "values": [2.0, 4.0, 4.0, 4.0, None, 2.0]},
         "flt": {"unit": "kg",
                 "values": [3.0, 39.0, None, None, 200.0, 9.0]},
         "log": {"unit": "onoff",
                 "values": [True, False, True, False, True, False]}
      },
      "destinations": {"your_farm": None, "my_farm": None, "farms_galore": None}
    }
    # fmt: on

    table_from_json = json_data_to_table(table_json_data, fixer=custom_test_fixer)
    assert table_from_cell_grid.equals(table_from_json)

    # Round trip
    table_json_data_back = table_to_json_data(table_from_json)
    table_from_json_round_trip = json_data_to_table(table_json_data_back, fixer=custom_test_fixer)
    assert table_from_cell_grid.equals(table_from_json_round_trip)


def test_fat():
    """ Factory Acceptance Test

        Read input files as dictionary objects.
        Compare objects to stored target objects (input_dir() / all.json)

    """
    all_files = 0
    for fn in os.listdir(input_dir()):
        path = input_dir() / fn
        if not os.path.isfile(path):
            continue
        if fn in ["auto_fixed.py", "__init__.py", "all.json", "all.csv"]:
            continue
        all_files += 1

    with open(input_dir() / "all.json") as f:
        all_json = json.load(f)

    count = 0
    for fn in os.listdir(input_dir()):
        path = input_dir() / fn
        if not os.path.isfile(path):
            continue
        if fn in ["auto_fixed.py", "__init__.py", "all.json", "all.csv"]:
            continue
        with open(input_dir() / fn, "r") as fh:
            cell_rows = (line.rstrip("\n").split(";") for line in fh)
            g = parse_blocks(
                cell_rows, **{"origin": f'"{fn}"', "to": "jsondata"}, fixer=custom_test_fixer
            )

            for tp, tt in g:
                if tp == BlockType.TABLE:
                    """  compare generic object
                         i.e. containing None instead of pd.NaT, np.nan &c.
                    """
                    count += 1
                    assert tt == all_json[fn]

    assert count == all_files


def test_to_json_serializable():
    """ Unit test pure_json_obj / json_esc
    """
    obj = {
        "k1": r'k1 w. \"quotes"',
        "nix": None,
        "no-flt": np.nan,
        "no-date": pd.NaT,
        "flt": 1.23,
        "int": 123,
    }
    js_obj = to_json_serializable(obj)

    # verify that js_obj is directly json serializable
    jstr = json.dumps(js_obj)
    js_obj_from_json = json.loads(jstr)

    assert js_obj_from_json["k1"] == obj["k1"]
    assert js_obj_from_json["nix"] is None
    assert js_obj_from_json["no-flt"] is None
    assert js_obj_from_json["no-date"] is None
    assert js_obj_from_json["flt"] == obj["flt"]
    assert js_obj_from_json["int"] == obj["int"]


def test_preserve_column_order():
    """ Unit test
        Verify that column order is preserved when translating btw. jsondata
        and pdtable.Table
    """
    # fmt: off
    lines_target = [
        ["**col_order"],
        ["dst2 dst2 dst2"],
        ["species" , "a3"  , "a2"  , "a1"  , "a4"  ],
        ["text"    , "-"   , "-"   , "-"   , "-"   ],
        ["chicken" , 1     , 2     , 3     , 4     ],
        ["pig"     , 1     , 2     , 3     , 4     ],
        ["goat"    , 1     , 2     , 3     , 4     ],
        ["zybra"   , 1     , 2     , 3     , 4     ],
        ["cow"     , 1     , 2     , 3     , 4     ],
        ["goose"   , 1     , 2     , 3     , 4     ],
    ]
    # fmt: on

    pandas_pdtab = make_table(lines_target)
    js_obj = table_to_json_data(pandas_pdtab)
    pdtab = json_data_to_table(js_obj)
    assert pdtab.df.iloc[0][3] == 3
    assert pdtab.df.iloc[1][1] == 1
    assert pdtab.df.iloc[2][4] == 4
    assert pandas_pdtab.equals(pdtab)

    # now verify from json-string
    jstr = json.dumps(js_obj)
    js_obj_from_json = json.loads(jstr)
    pdtab_from_json = json_data_to_table(js_obj_from_json)

    assert pandas_pdtab.equals(pdtab_from_json)
